@@ -11,7 +11,7 @@ use JSON::XS qw(decode_json encode_json);
 use Time::HiRes qw(gettimeofday time);
 use Data::Dumper;
 
-our $VERSION = "0.011";
+our $VERSION = "0.020";
 
 my %INFO_ROUTES = (
     version => '/version',
@@ -68,10 +68,10 @@ sub list {
     });
 
     if ( ! $rs->{success} ) {
-       $self->{last_error} = _extract_error_tiny($rs); 
+       $self->{last_error} = _extract_error_tiny($uri, $rs);
        return undef;
     }
- 
+
     my $response = decode_json($rs->{content});
 
     my @tables = ();
@@ -103,7 +103,7 @@ sub version {
     });
 
     if ( ! $rs->{success} ) {
-       $self->{last_error} = _extract_error_tiny($rs); 
+       $self->{last_error} = _extract_error_tiny($uri, $rs);
        return undef;
     }
 
@@ -119,10 +119,19 @@ sub version {
 # usage:
 #
 # my $records = $hbase->get({
-#    table       => $table_name,
-#    where       => {
-#        key_begins_with => "$key_prefix"
-#    },
+#   table => $table_name,
+#   where => {
+#        key_equals => $key
+#   },
+#   columns => [
+#       'd:some_column_name',
+#       'd:some_other_column_name'
+#   ],
+#   versions => 100,
+#   timestamp_range => {
+#       from  => $timestamp_from,
+#       until => $timestamp_until,
+#   }
 # });
 #
 sub get {
@@ -131,7 +140,9 @@ sub get {
 
     $self->{last_error} = undef;
 
-    my $rows = $self->_get_tiny($params);
+    my $uri = _build_get_uri($params);
+
+    my $rows = $self->_get_tiny($uri);
 
     return $rows;
 }
@@ -139,33 +150,21 @@ sub get {
 # _get_tiny
 sub _get_tiny {
 
-    my $self  = shift;
-    my $query = shift;
+    my $self = shift;
+    my $uri  = shift;
 
-    my $table = $query->{table};
-
-    my $route;
-    if ($query->{where}->{key_equals}) {
-        my $key = $query->{where}->{key_equals};
-        $route = '/' . $table . '/' . uri_escape($key);
-    }
-    else {
-        my $part_of_key = $query->{where}->{key_begins_with};
-        $route = '/' . $table . '/' . uri_escape($part_of_key . '*');
-    }
-
-    my $uri = $self->{service} . $route;
+    my $url = $self->{service} . $uri;
 
     my $http = HTTP::Tiny->new();
 
-    my $rs = $http->get($uri, {
+    my $rs = $http->get($url, {
         headers => {
             'Accept' => 'application/json',
         }
     });
 
     if ( ! $rs->{success} ) {
-       $self->{last_error} = _extract_error_tiny($uri, $rs);
+       $self->{last_error} = _extract_error_tiny($url, $rs);
        return undef;
     }
 
@@ -193,40 +192,36 @@ sub _get_tiny {
 #
 # multiget
 #
+# usage:
+#
+# my $records = $hbase->multiget({
+#    table   => $table_name,
+#    where   => {
+#        key_in => \@keys
+#    },
+#    versions => $number_of_versions,
+# });
+#
 sub multiget {
     my $self  = shift;
     my $query = shift;
 
     $self->{last_error} = undef;
 
-    my $keys  = $query->{where}->{key_in};
-    my $table = $query->{table};
-
-    my $route_base = '/' . $table . '/multiget?';
-    my $uri_base = $self->{service} . $route_base;
-
-    my @multiget_urls = ();
-    my $current_url = undef;
-    foreach my $key (@$keys) {
-        if (! defined $current_url) {
-            $current_url ||= $uri_base . "row=" . $key;
-        }
-        else{
-            my $next_url = $current_url . '&row=' . $key;
-            if (length($next_url) < 2000) {
-                $current_url = $next_url;
-            }
-            else {
-                push @multiget_urls, { url => $current_url, len => length($current_url) };
-                $current_url = undef;
-            }
-        }
+    my $where = $query->{where};
+    unless ($where->{key_in} && @{$where->{key_in}}) {
+        $self->{last_error} = {
+            type => "Bad request",
+            info => "No keys specified for multiget.",
+            uri  => "Could not counstruct uri - no keys provided.",
+        };
+        return;
     }
-    # last batch
-    push @multiget_urls, { url => $current_url, len => length($current_url) };
+
+    my $multiget_urls = _build_multiget_uri($query);
 
     my @result = ();
-    foreach my $url (@multiget_urls) {
+    foreach my $url (@$multiget_urls) {
         my $rows = $self->_multiget_tiny($url->{url});
         push @result, @$rows
             if ($rows and @$rows);
@@ -244,18 +239,20 @@ sub _multiget_tiny {
     my $self = shift; # hbh
     my $uri  = shift;
 
+    my $url = $self->{service} . $uri;
+
     my $http = HTTP::Tiny->new();
 
     my $data_format = 'application/json';
 
-    my $rs = $http->get($uri, {
+    my $rs = $http->get($url, {
         headers => {
             'Accept' => $data_format,
         }
     });
 
     if ( ! $rs->{success} ) {
-       $self->{last_error} = _extract_error_tiny($rs);
+       $self->{last_error} = _extract_error_tiny($url, $rs);
        return undef;
     }
 
@@ -290,9 +287,13 @@ sub _multiget_tiny {
 #       {
 #          row_key   => "$row_key",
 #          row_cell => [
-#              { column => 'family:name', value => 'value' },
+#              {
+#                   column    => "$family:$name",
+#                   value     => "$value",
+#                   timestamp => $timestamp # <- optional (override HBase timestamp) 
+#               },
 #              ...,
-#              { column => 'family:name', value => 'value' },
+#              { column => "$family:$name", value => "$value" },
 #         ],
 #      },
 #      ...
@@ -340,8 +341,6 @@ sub put {
         my $row_change_formated = { Row => $rows };
         my $row_cell_changes_formated = {};
 
-        my $ts = int(gettimeofday * 1000);
-
         # hbase wants keys in sorted order; it wont work otherwise;
         # more specificaly, the special key '$' has to be at the end;
         my $sorted_json_row_change =
@@ -353,11 +352,22 @@ sub put {
         my @sorted_json_cell_changes = ();
         foreach my $cell_change (@$row_cell_changes) {
 
+            # timestamp override
+            my $ts;
+            my $overide_timestamp = undef;
+            if ($cell_change->{timestamp}) {
+                $overide_timestamp = 1;
+                $ts = $cell_change->{timestamp};
+            }
+
+            my $timestamp_override_string = $overide_timestamp
+                ? '"timestamp":"' . $ts . '",'
+                : ''
+            ;
+
             my  $sorted_json_cell_change =
                     '{'
-                        . '"timestamp":"'
-                        . $ts
-                        . '",'
+                        . $timestamp_override_string
                         . '"column":"'
                         . encode_base64($cell_change->{column}, '')
                         . '",'
@@ -394,22 +404,190 @@ sub put {
     });
 
     if ( ! $rs->{success} ) {
-       $self->{last_error} = _extract_error_tiny($rs); 
+       $self->{last_error} = _extract_error_tiny($uri, $rs);
        return undef;
     }
- 
+
+}
+
+# =========================================================================
+# delete: delete an entire record or selected columns of it
+#
+# Usage:
+# my $success = $hbh->delete({
+#   table    => 'table',
+#   key      => 'key',
+#   family   => 'family', # optional, unless column is given
+#   column   => 'column', # optional
+# })
+sub delete {
+    my ($self, $attr) = @_;
+    my $table  = delete $attr->{table};
+    my $key    = delete $attr->{key};
+    my $family = delete $attr->{family};
+    my $column = delete $attr->{column};
+    my ($route, $http, $rs, $url);
+
+    die "Table name required" if(!$table);
+    die "Row key required" if(!$key);
+    die "Family is required if column is given" if($column && !$family);
+
+    $key = join(';', @$key) if(ref($key) eq 'ARRAY');
+    $route = sprintf("/%s/%s", $table, uri_escape($key));
+
+    if($family) {
+        $route .= sprintf("/%s", $family);
+        $route .= sprintf(":%s", $column) if($column);
+    }
+
+    $http = HTTP::Tiny->new();
+    $url = sprintf("%s%s", $self->{service}, $route);
+    $rs = $http->delete($url, {
+        headers => {
+            'Accept' => 'application/json',
+        }
+    });
+
+    if($rs->{success}) {
+       $self->{last_error} = undef;
+       return 1;
+    }
+    else {
+        $self->{last_error} = _extract_error_tiny($url, $rs);
+        return 0;
+    }
+}
+
+# -------------------------------------------------------------------------
+# build get uri
+#
+sub _build_get_uri {
+    my $query = shift;
+
+    my $table = $query->{table};
+
+    my $columns_url_part = "";
+    if ( $query->{columns} and @{$query->{columns}} ) {
+        $columns_url_part = "/" . join(",", map { uri_escape($_) } @{ $query->{columns} });
+    }
+
+    my $timestamp_url_part  = undef;
+    # timestamp range query is supported only if columns are specifed
+    if ($query->{columns} and @{$query->{columns}}) {
+        if ( $query->{timestamp_range} and %{ $query->{timestamp_range} } ) {
+            my $timestamp_from  = $query->{timestamp_range}->{from};
+            my $timestamp_until = $query->{timestamp_range}->{until};
+            $timestamp_url_part = "/" . $timestamp_from . "," . $timestamp_until;
+        }
+    }
+
+    my $versions_url_part = undef;
+    if ( $query->{versions} ) {
+        my $versions = $query->{versions};
+        $versions_url_part = "?v=$versions";
+    }
+
+    my $uri;
+    if ($query->{where}->{key_equals}) {
+        my $key = $query->{where}->{key_equals};
+
+        $uri = '/' . $table . '/' . uri_escape($key) . $columns_url_part;
+
+        $uri .= $timestamp_url_part if $timestamp_url_part;
+        $uri .= $versions_url_part if $versions_url_part;
+    }
+    else {
+        my $part_of_key = $query->{where}->{key_begins_with};
+        $uri = '/' . $table . '/' . uri_escape($part_of_key . '*') . $columns_url_part;
+
+        $uri .= $timestamp_url_part if $timestamp_url_part;
+        $uri .= $versions_url_part if $versions_url_part;
+    }
+
+    return $uri;
+}
+
+# -------------------------------------------------------------------------
+# build multiget url
+#
+sub _build_multiget_uri {
+    my $query = shift;
+
+    my $keys  = $query->{where}->{key_in};
+    my $table = $query->{table};
+
+    my $route_base = '/' . $table . '/multiget?';
+    my $uri_base = $route_base;
+
+    my @multiget_urls = ();
+    my $current_url = undef;
+    foreach my $key (@$keys) {
+        if (! defined $current_url) {
+            $current_url ||= $uri_base . "row=" . $key;
+        }
+        else{
+            my $next_url = $current_url . '&row=' . $key;
+            if (length($next_url) < 2000) {
+                $current_url = $next_url;
+            }
+            else {
+                push @multiget_urls, { url => $current_url, len => length($current_url) };
+                $current_url = undef;
+            }
+        }
+    }
+    # last batch
+    push @multiget_urls, { url => $current_url, len => length($current_url) };
+
+    if ($query->{versions}) {
+        foreach my $mget_url (@multiget_urls) {
+            my $versions = $query->{versions};
+            my $versions_url_part = "v=$versions";
+            
+            $mget_url->{url} .= '&' . $versions_url_part;
+        }
+    }
+
+    return \@multiget_urls;
 }
 
 # -------------------------------------------------------------------------
 # parse error
 #
 sub _extract_error_tiny {
+
     my $uri = shift;
     my $res = shift;
 
     return if $res->{success};
-    return if $res->{status} == 404;
-    my $msg = $res->{reason};
+
+    if (!$res->{status}) {
+         return {
+            type => 'Unkown',
+            info => 'No status in response',
+            uri => $uri
+        };
+    }
+
+    if ($res->{status} and $res->{status} == 404) {
+        return {
+            type => '404',
+            info => 'No rows found',
+            uri => $uri
+        };
+    }
+
+    my $msg;
+    if ($res->{reason}) {
+        $msg = $res->{reason};
+    }
+    else {
+        return {
+            type => "Bad response",
+            info => "No reason in the response",
+            uri  => $uri
+        };
+    }
 
     my ($exception, $info) = $msg =~ m{\.([^\.]+):(.*)$};
     if ($exception) {
@@ -498,7 +676,7 @@ Scans a table by key prefix or exact key match depending on options passed:
         },
     });
 
-    # exact match:
+    # exact key match: get the whole row
     my $record = $hbase->get({
         table       => $table_name,
         where       => {
@@ -506,6 +684,47 @@ Scans a table by key prefix or exact key match depending on options passed:
         },
     });
 
+    # exact key match: get only specific columns
+    my $record = $hbase->get({
+      table => $table_name,
+      where => {
+           key_equals => $key
+      },
+      columns => [
+          'd:some_column_name',
+          'd:some_other_column_name'
+      ],
+    });
+
+    # exact key match: get last $N cell versions
+    my $records = $hbase->get({
+      table => $table_name,
+      where => {
+           key_equals => $key
+      },
+      columns => [
+          'd:some_column_name',
+          'd:some_other_column_name'
+      ],
+      versions => $N,
+    });
+
+    # exact key match: get cell versions created within a timestamp range
+    my $records = $hbase->get({
+      table => $table_name,
+      where => {
+           key_equals => $key
+      },
+      columns => [
+          'd:some_column_name',
+          'd:some_other_column_name'
+      ],
+      timestamp_range => {
+          from  => $timestamp_from,
+          until => $timestamp_until,
+      }
+    });
+
 =head2 multiget
 
 Does a multiget request to HBase, so that multiple keys can be matched
@@ -514,8 +733,8 @@ than 2000 chars, so if the number of keys passed is large enough and would
 result in url longer than 2000 chars, the request is split into multiple
 smaller request so each is shorter than 2000 chars.
 
+    # multiget: get only last cell version from matched rows
     my @keys = ($key1,...,$keyN);
-
     my $records = $hbase->multiget({
         table   => $table_name,
         where   => {
@@ -523,11 +742,22 @@ smaller request so each is shorter than 2000 chars.
         },
     });
 
+    # multiget: get last $N cell versions from matched rows
+    my @keys = ($key1,...,$keyN);
+    my $records = $hbase->multiget({
+        table   => $table_name,
+        where   => {
+            key_in => \@keys
+        },
+        versions => $N,
+    });
+
 =head2 put
 
 Inserts one or multiple rows. If a key allready exists then depending
-on if HBase versioning is on, the record will be updated (versioning is off)
-or new version will be inserted (versioning is on)
+on if HBase versioning is ON for that specific table, the record will
+be updated (versioning is off) or new version will be inserted (versioning
+is on)
 
     # multiple rows
     my $rows = [
@@ -537,7 +767,17 @@ or new version will be inserted (versioning is on)
 
             # cells: array of hashes where eash hash is one cell
             row_cells => [
-              { column => "$family_name:$colum_name", value => "$value" },
+                {
+                    column => "$family_name1:$colum_name1",
+                    value  => "$value1",
+                    timestamp => "$timestamp1", # <- optional (override HBase timestamp)
+                },
+                ...,
+                {
+                    column => "$family_nameN:$colum_nameN",
+                    value  => "$valueN",
+                    timestamp => "$timestampN", # <- optional (override HBase timestamp)
+                },
             ],
        },
        ...
@@ -566,6 +806,17 @@ or new version will be inserted (versioning is on)
         changes => $rows
     });
 
+=head2 delete
+
+Deletes an entire record or selected columns of it
+
+    my $success = $hbase->delete({
+        table    => 'table',
+        key      => 'key',
+        family   => 'family', # optional, unless column is given
+        column   => 'column', # optional
+    });
+
 =head2 version
 
 Returns a hashref with server version info
@@ -582,7 +833,7 @@ Returns a hashref with server version info
           'OS' => 'Linux 2.6.32-358.23.2.el6.x86_64 amd64',
           'JVM' => 'Oracle Corporation 1.7.0_51-24.51-b03'
         };
-    
+
 =head2 list
 
 Returns a list of tables available in HBase
@@ -618,7 +869,7 @@ Information on error is stored in hbase object under key last error:
 
 =head1 VERSION
 
-Current version: 0.011
+Current version: 0.020
 
 =head1 AUTHOR
 
@@ -632,6 +883,8 @@ Sawyer X, C<< <xsawyerx at cpan.org> >>
 
 Eric Herman, C<< <eherman at cpan.org> >>
 
+Robert Nilsson, <rn@orbstation.com>
+
 =head1 COPYRIGHT
 
 Copyright (c) 2014 the HBase::JSONRest L</AUTHOR> and L</CONTRIBUTORS>
